using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MeshVR;
using SimpleJSON;

namespace EmissiveClothing
{
    public class EmissiveClothing : MVRScript
    {

		JSONStorableFloat alpha;
        JSONStorableBool renderOriginal;
        JSONStorableUrl loadedShaderPath = new JSONStorableUrl("shader", "");
		JSONStorableBool originalColor;
		JSONStorableBool mode;
        bool loadedAssetBundle;
        bool isLoading;
        int retryAttempts;
		bool clearEmissionMat = false;

        private static readonly string BUNDLE_NAME = "emissiveshader.assetbundle";

        protected List<Material[]> ourMaterials = new List<Material[]>();
		protected List<Material[]> ourMaterialsBackup = new List<Material[]>();
        protected Shader shader;
        protected List<DAZSkinWrap> wraps = new List<DAZSkinWrap>();
        protected JSONStorableColor color;

		
		public override void Init()
        {
            try {
                color = new JSONStorableColor("Color", HSVColorPicker.RGBToHSV(1f, 1f, 1f), _ => SyncMats());
                RegisterColor(color);
                CreateColorPicker(color, true);
                FloatSlider(ref alpha, "Color Alpha", 1,
                    _ => { SyncMats(); }, 0, 2, true);

                renderOriginal = new JSONStorableBool("Render Original Material", true);
                RegisterBool(renderOriginal);
                CreateToggle(renderOriginal, false);
				
				originalColor = new JSONStorableBool("Use Original Color", true, (bool _) => SyncMats());
                CreateToggle(originalColor, false);

                mode = new JSONStorableBool("Mode", true, (bool _) => SyncMats());
                CreateToggle(mode, false);

                RegisterUrl(loadedShaderPath);

                CreateButton("Rescan active clothes").button.onClick.AddListener(() => {
                    StartCoroutine(Rebuild());
                    });
                
                StartCoroutine(LoadShaderAndInit());
            } catch (Exception e) {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        
        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            isLoading = true;
            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
        }

        string GetPluginPath() // basically straight from VAMDeluxe's Dollmaster
        {
            string pluginId = this.storeId.Split('_')[0];
            string pathToScriptFile = manager.GetJSON(true, true)["plugins"][pluginId].Value;
            string pathToScriptFolder = pathToScriptFile.Substring(0, pathToScriptFile.LastIndexOfAny(new char[] { '/', '\\' }));
            return pathToScriptFolder;
        }

        private void SyncMats()
        {
			                ////SuperController.LogMessage("SyncMats Starting");
			int i = 0;
			
            foreach (var r in ourMaterials) {
				var msb = ourMaterialsBackup[i];
				int j =0;
                foreach (var m in r) {
					var mb = msb[j];
                    if (m == null){
						j += 1;
                        continue;
					}
                   /*  Color c = color.colorPicker.currentColor;
					if(c!=Color.white)
                    m.SetColor("_Color", c);
				
				
					if(alpha.val!=0f){
					c = m.GetColor("_Color");
					c.a = alpha.val;
					m.SetColor("_Color", c);
					}
					if(clearEmissionMat==true){
						SuperController.LogMessage("Clearing Emission Material");
						m.SetColor("_Color", Color.black);
						m.SetFloat("_AlphaAdjust", -1f);
						renderOriginal.val = false;
					}
					 */
					Color c = color.colorPicker.currentColor;
					if(c!=Color.white)
                    m.SetColor("_Color", c);
				
					if(alpha.val!=0f){
						if (mode.val){
							if (originalColor.val){
							c = mb.GetColor("_Color");}
							else{
							c = color.colorPicker.currentColor;}
							c.a = alpha.val;
						}
						else{
							if (originalColor.val){
							c = mb.GetColor("_Color");}
							else{
							c = color.colorPicker.currentColor;}
							c *= alpha.val;
						}
                    m.SetColor("_Color", c);
                    
					
					}
					if(clearEmissionMat==true){
						SuperController.LogMessage("Clearing Emission Material");
						m.SetColor("_Color", Color.black);
						m.SetFloat("_AlphaAdjust", -1f);
						renderOriginal.val = false;
						//clearEmissionMat==false;
					}
					
					j += 1;
                }
				i += 1;
            }
        }

        private void FloatSlider(ref JSONStorableFloat output, string name, float start, JSONStorableFloat.SetFloatCallback callback, float min, float max, bool rhs)
        {
            output = new JSONStorableFloat(name, start, callback, min, max, false, true);
            RegisterFloat(output);
            CreateSlider(output, rhs);
        }

        // MeshVR.AssetLoader provides some nice refcounting for us.
        // The downsides are the annoying callback API, that we can't suppress error messages, and that a refcount really shouldn't be created on failure, wat
        class LoadAssetBundle : CustomYieldInstruction {
            AssetLoader.AssetBundleFromFileRequest req;

            private bool done;

            // We can't write the type AssetBundle, but we can call methods on it. lolwut C#
            public bool success => req.assetBundle != null;
            public T GetAsset<T>(string name) where T : UnityEngine.Object
            {
                return req.assetBundle?.LoadAsset<T>(name);
            }

            public LoadAssetBundle(string url)
            {
                req = new AssetLoader.AssetBundleFromFileRequest {
                    path = url,
                    callback = _ => {
                        done = true;
                        if (req.assetBundle == null)
                            AssetLoader.DoneWithAssetBundleFromFile(req.path);
                    }
                };
                AssetLoader.QueueLoadAssetBundleFromFile(req);
            }

            public override bool keepWaiting => !done;
        }

        private IEnumerator AttemptLoadShader(string url)
        {
            if (shader != null)
                yield break;
            var request = new LoadAssetBundle(url);
            yield return StartCoroutine(request);

            if (!request.success)
                yield break;
            shader = request.GetAsset<Shader>("Assets/EmissiveHack.shader");
            if (shader != null) {
                loadedShaderPath.val = url;
                loadedAssetBundle = true;
            } else {
                SuperController.LogError("Bad emissiveshader assetbundle");
                AssetLoader.DoneWithAssetBundleFromFile(url);
            }
        }

        // If there are multiple copies of this script loaded from different paths, avoid failures due to trying to load the same assetbundle. 
        // This is particularly important because sometimes the bundle seems to get leaked and not unload the shader.
        // Stash the url as well so that vac saves work properly even if this happens
        /* private static readonly string STASH_NAME = "Alazi.EmissiveShaderStash";
        private void StashShaderName()
        {
            PlayerPrefs.SetString(STASH_NAME, loadedShaderPath.val);
        }
        private void LoadStashedShader()
        {
            shader = Resources.FindObjectsOfTypeAll<Shader>().Where(s => s.name == "Custom/Alazi/ExtraEmissiveComputeBuff").FirstOrDefault();
            if (shader != null) {
                loadedShaderPath.val = PlayerPrefs.GetString(STASH_NAME);
            } else {
                PlayerPrefs.DeleteKey(STASH_NAME);
            }
        } */

        protected IEnumerator LoadShaderAndInit()
        {
			SuperController.LogMessage("LoadShaderAndInit");
            // Wait until json load is done
            yield return new WaitForEndOfFrame();
            //LoadStashedShader();
			SuperController.LogMessage("LoadStashedShader");
            if (shader == null) {
                if (loadedShaderPath.val != "")
                    yield return AttemptLoadShader(loadedShaderPath.val);
                yield return AttemptLoadShader($"{GetPluginPath()}/{BUNDLE_NAME}");
                yield return AttemptLoadShader($"Custom/Assets/{BUNDLE_NAME}");

                if (shader == null) {
                    SuperController.LogError("Failed to load shader");
                    yield break;
                }
                //StashShaderName();
            }

            Build();
        }

        protected IEnumerator Rebuild()
        {
									SuperController.LogMessage("Rebuild");
            Unbuild();
            // wait for components to be destroyed
            yield return new WaitForEndOfFrame();
            Build();
        }
        protected void Build()
        {
						SuperController.LogMessage("Build");
            var allWraps = containingAtom.gameObject.GetComponentsInChildren<DAZSkinWrap>(false);
            ourMaterials = new List<Material[]>();
			ourMaterialsBackup = new List<Material[]>();
            wraps = new List<DAZSkinWrap>();
            if (allWraps.Length == 0) {
                SuperController.LogMessage("No clothes loaded");
                return;
            }
            bool doRescan = false;

            foreach (var wrap in allWraps) {
				    SuperController.LogMessage("wrap="+wrap.ToString());
                if (wrap.ToString().Contains("Emissive")) {
                    SuperController.LogError($"EmissiveClothing: found dup {wrap}");
                    continue;
                }
                if (wrap.skin.delayDisplayOneFrame) {
                    SuperController.LogError($"EmissiveClothing: {wrap} is delayed, not set up to handle that");
                    continue;
                }
                var ourMats = new Material[wrap.GPUmaterials.Length];
				var ourMatsBackup = new Material[wrap.GPUmaterials.Length];
                var theirNewMats = wrap.GPUmaterials.ToArray();
                bool foundAny = false;
                
                foreach (var mo in wrap.GetComponents<DAZSkinWrapMaterialOptions>()) {
					SuperController.LogMessage("wrap.GetComponents<DAZSkinWrapMaterialOptions>.overrideId="+mo.overrideId);
                    if (!mo.overrideId.Contains("(em)"))
                        continue;
                    // too lazy to duplicate all the code for slots2 / simpleMaterial
                    if (mo.paramMaterialSlots?.Length == 0) 
                        continue;
                    foundAny = true;
                    
					
					
                    foreach (var i in mo.paramMaterialSlots) {
                        var mat = wrap.GPUmaterials[i];
						SuperController.LogMessage("GPUmaterials Length="+wrap.GPUmaterials.Length);
						//SuperController.LogError("GPUmaterials Length="+wrap.GPUmaterials.Length);
                        var ourMat = new Material(shader);
						var ourMatBackup = new Material(shader);
                        ourMats[i] = ourMat;
						ourMatsBackup[i] = ourMatBackup;
						
						//SuperController.LogError("Reload Error Point Next "+mat.GetType()+mat.name);						
                        ourMat.name = mat.name;
						ourMatBackup.name = mat.name;
						//SuperController.LogError("Reload Error Point Passed");
						//SuperController.LogError("ourMat name="+mat.name);
                        // Ideally we'd hook all the config stuff in MaterialOptions, but that would 
                        // require too much effort to reimplement all the url/tile/offset->texture code 
                        // or to copy the existing one to override the relevant methods
                        // So require the user to hit rescan manually.
                        var tex = mat.GetTexture("_DecalTex");
                        ourMat.SetTexture("_MainTex", tex);
						
						Color col=mat.GetColor("_Color");
						ourMat.SetColor("_Color", mat.GetColor("_Color"));
						ourMatBackup.SetColor("_Color", mat.GetColor("_Color"));
						
						if (ourMat.GetColor("_Color")==Color.black)
							ourMat.SetColor("_Color", Color.white);
						
						ourMat.SetFloat("_AlphaAdjust", mat.GetFloat("_AlphaAdjust"));
						//SuperController.LogError("AlphaSRC_ "+ourMat.GetFloat("_AlphaAdjust")+" Color "+ourMat.GetColor("_Color"));
						
                        mat.SetTexture("_DecalTex", null);

                        // Particularly during scene load it may take a while for these to be loaded, try a few times
                        if (tex == null)
						{
							                SuperController.LogMessage("1-Texture is NULL/Not Loaded in Build/Doing Rescaning");
						doRescan = true;
						}
						SuperController.LogMessage("1.5-Texture Maybe Loaded in Build/Doing Copy");
                        // could maybe get some tiny extra performance by using a null shader instead
                        theirNewMats[i] = new Material(mat);
                        theirNewMats[i].SetFloat("_AlphaAdjust", -1); // Hidden Toggle Alpha set to 0
												if (doRescan == true)
												SuperController.LogMessage("Build Failed");
											else SuperController.LogMessage("Build Finished");
											
                    }
                }
                if (!foundAny)
                    continue;

                ourMaterials.Add(ourMats);
				ourMaterialsBackup.Add(ourMatsBackup);
				
                wrap.BroadcastMessage("OnApplicationFocus", true);
				SuperController.LogMessage("wrap.gameObject.AddComponent<EmissiveDAZSkinWrap>");
                wrap.gameObject.AddComponent<EmissiveDAZSkinWrap>().CopyFrom(wrap, theirNewMats, ourMats, renderOriginal);
				SuperController.LogMessage("wraps.Add(wrap)");
                wraps.Add(wrap);
            }

            SyncMats();
            if (doRescan)
                StartCoroutine(QueueRescan());
            else
                isLoading = false;
        }

        private IEnumerator QueueRescan()
        {
			SuperController.LogMessage("2-QueueRescaning-Return Rebuild");
            // 1s normally or 40s during load
            if (retryAttempts < 5 || (isLoading && retryAttempts < 200000)) {
                //yield return new WaitForSeconds(0.2f);
				SuperController.LogMessage("3-QueueResReAttempt:"+retryAttempts);
                retryAttempts++;
                yield return Rebuild();
            }

        }

        void Unbuild()
        {
			/* SuperController.LogMessage("Unbuild");
			SuperController.LogMessage("wraps.Count="+wraps.Count+" "+wraps.GetType());
            for (int i = 0; i < wraps.Count; i++) {
                var wrap = wraps[i];
                var mats = ourMaterials[i];
				var matsBackup = ourMaterialsBackup[i];
				SuperController.LogMessage("mats.Length "+mats.Length+" wraps[i]="+wraps[i]);
                for (int j = 0; j < mats.Length; j++) {
                    if (mats[j] == null)
                        continue;
                    // If it's been changed, don't reset it
                    //if (wrap.GPUmaterials[j].GetTexture("_DecalTex") == null){
						//SuperController.LogMessage("DecalTex is Null");
                        wrap.GPUmaterials[j].SetTexture("_MainTex", matsBackup[j].GetTexture("_MainTex"));
						wrap.GPUmaterials[j].SetTexture("_DecalTex", matsBackup[j].GetTexture("_DecalTex"));
						wrap.GPUmaterials[j].SetColor("_Color", matsBackup[j].GetColor("_Color"));
                }
				//SuperController.LogMessage("Now Destroying wrap.gameObject EmissiveDAZSkinWrap");
				SuperController.LogMessage("Now Destroying = "+wrap.gameObject?.GetComponent<EmissiveDAZSkinWrap>().name);
				
				
				 foreach (var Compo in wrap.gameObject?.GetComponents<EmissiveDAZSkinWrap>()) {
					 SuperController.LogMessage("Components Found= "+Compo.name+" Type= "+Compo.GetType());
				 }
				
				
				
                GameObject.Destroy(wrap.gameObject.GetComponent<EmissiveDAZSkinWrap>());
                //wrap.draw = true;
                var control = wrap.gameObject?.GetComponent<DAZSkinWrapControl>();
				//SuperController.LogMessage("Now control = "+wrap.gameObject?.GetComponent<DAZSkinWrapControl>().name);
                if (control && (control.wrap == null || control.wrap == this)) {
					//SuperController.LogMessage("Now controling");
                  control.wrap = wrap;  
                }
				
            } */
			
			for (int i = 0; i < wraps.Count; i++) {
                var wrap = wraps[i];
                var mats = ourMaterials[i];
				var matsBackup = ourMaterialsBackup[i];
                for (int j = 0; j < mats.Length; j++) {
                    if (mats[j] == null)
                        continue;
					if (wrap.GPUmaterials[j].GetTexture("_DecalTex") == null){
                    wrap.GPUmaterials[j].SetTexture("_MainTex", matsBackup[j].GetTexture("_MainTex"));
					wrap.GPUmaterials[j].SetTexture("_DecalTex", matsBackup[j].GetTexture("_DecalTex"));
					wrap.GPUmaterials[j].SetColor("_Color", matsBackup[j].GetColor("_Color"));
					}
                }

                GameObject.Destroy(wrap.gameObject?.GetComponent<EmissiveDAZSkinWrap>());
                wrap.draw = true;
                var control = wrap.gameObject?.GetComponent<DAZSkinWrapControl>();
                if (control && (control.wrap == null || control.wrap == this)) {
                    control.wrap = wrap;
                }
            }
			//clearEmissionMat = true;
        }

        void OnDestroy()
        {
						SuperController.LogMessage("OnDestroy Emission Will Not Turn Off...Next Disable/Force Reload Cloth");
						Unbuild();
						clearEmissionMat = true;
						SyncMats();

            if (loadedAssetBundle) {
                //AssetLoader.DoneWithAssetBundleFromFile(loadedShaderPath.val);
            }
        }
		
		IEnumerator waiter()
{
	SuperController.LogError("Wait for 4 seconds");
    //Wait for 4 seconds
    yield return new WaitForSeconds(4);
	SuperController.LogError("Wait Finished 2");
}
		

        // The only thing we absolutely need and can't get normally from 
        // DAZSkinWrap is the ComputeBuffers; Material.GetBuffer does not exist
        class EmissiveDAZSkinWrap : DAZSkinWrap
        {
            Dictionary<int, ComputeBuffer> emissiveMaterialVertsBuffers = new Dictionary<int, ComputeBuffer>();

            JSONStorableBool renderOriginal;
            bool oldOriginal;
            Material[] hiddenMats;
            Material[] emissiveMats;

            public void CopyFrom(DAZSkinWrap wrap, Material[] hiddenMats, Material[] emissiveMats, JSONStorableBool renderOriginal)
            {
                base.skinTransform = wrap.skinTransform;
                base.skin = wrap.skin;
                base.dazMesh = wrap.dazMesh;
                base.GPUSkinWrapper = wrap.GPUSkinWrapper;
                base.GPUMeshCompute = wrap.GPUMeshCompute;
                base.CopyMaterials();
                base.GPUmaterials = wrap.GPUmaterials;
                this.renderOriginal = renderOriginal;
                this.hiddenMats = hiddenMats;
                this.emissiveMats = emissiveMats;

                base.wrapName = wrap.wrapName;
                base.wrapStore = wrap.wrapStore;

                wrap.draw = false;
                base.draw = true;

                base.surfaceOffset = wrap.surfaceOffset;
                base.defaultSurfaceOffset = wrap.defaultSurfaceOffset;
                base.additionalThicknessMultiplier = wrap.additionalThicknessMultiplier;
                base.defaultAdditionalThicknessMultiplier = wrap.defaultAdditionalThicknessMultiplier;

                var control = GetComponent<DAZSkinWrapControl>();
                if (control && (control.wrap == null || control.wrap == wrap)) {
                    control.wrap = this;
                }
            }


            void LateUpdate() // Overwrite base one to possibly not render some mats
            {
                UpdateVertsGPU();
						//SuperController.LogMessage("renderOriginal="+renderOriginal.val+"   oldOriginal="+oldOriginal);				
                if (renderOriginal.val != oldOriginal) {					
                    materialVertsBuffers.Clear();
                    materialNormalsBuffers.Clear();
                    materialTangentsBuffers.Clear();
                    oldOriginal = renderOriginal.val;
                }
                if (!renderOriginal.val) {
									//SuperController.LogMessage("hiddenMats DrawMeshGPU()");
                    var temp = base.GPUmaterials;
                    base.GPUmaterials = hiddenMats;
                    DrawMeshGPU();
					//SuperController.LogMessage("hiddenMats DrawMeshGPU()"+renderOriginal.val);
                    base.GPUmaterials = temp;
                } else {
                    DrawMeshGPU();
                }
                
                for (int i = 0; i < emissiveMats.Length; i++) {
                    if (emissiveMats[i] == null)
                        continue;
                    emissiveMats[i].renderQueue = base.GPUmaterials[i].renderQueue; // could probably get away without updating this past initialization
                    UpdateBuffer(emissiveMaterialVertsBuffers, base.materialVertsBuffers, "verts", i);

                    Graphics.DrawMesh(base.mesh, Matrix4x4.identity, emissiveMats[i], 0, null, i, null, false, false);
                }
            }

            private void UpdateBuffer(Dictionary<int, ComputeBuffer> ours, Dictionary<int, ComputeBuffer> theirs, string buf, int i)
            {
                ComputeBuffer ourB, theirB;
                ours.TryGetValue(i, out ourB);
                theirs.TryGetValue(i, out theirB);
                if (ourB != theirB) {
                    emissiveMats[i].SetBuffer(buf, theirB);
                    ours[i] = theirB;
                }
            }
        }
    }

}
